/***********************************************************************/
/*                                                                     */
/*  File : RSA.java                                                   */
/*  Author : WEI-NING LIN                                              */
/*  Author : Isao yanagimachi                                          */
/*                                                                     */
/*  Description:                                                       */
/*   The data type long is used to store the encrypted message.        */
/*    Thus, maximum number of characters that can be encrypted is 4   */
/*    (long is four bytes) since we use the ascii character set (128) */
/*********************************************************************/

/*********NOTE ***************************/
// we found out that Java has some problem when handling the big number 
// even though it is within its type's precision.
// Also BigDecimal class does not work properly when it comes to dealing
// with the big number. Thus the value of  q*p  is not infinite.



import java.*;
import java.math.BigDecimal;

public class RSA 
{

   //finds the greated common divisor to find the e
  public long GCD(long e, long phi)
  {
       if ( phi == 0 ) 
         return e;
       else
         return GCD(phi, e % phi);
  }

  //finds the e, for public key
  public int findE(long p, long q)
  {
    long gcd = 0;
     int e = 2; 
     long phi = ( p-1)*(q-1);
     while ( gcd != 1 )
     {
        e = e+1;
        gcd = GCD(e,phi);
     }
     return e;
  }

  //use the extended euclien algorithm to find the D
  public long extend(long E, long PHI)
  {
     long u1 = 1;
     long u2 = 0;
     long u3 = PHI;
     long v1 = 0;
     long v2 = 1;
     long v3 = E;    
     long q;
     long t1;
     long t2;
     long t3;
     long z;
     long uu;
     long vv;
     long inverse;

     while (v3 != 0) {
        q = (long)Math.ceil(u3/v3);
        t1 = u1 - q * v1;
        t2 = u2 - q * v2;
        t3 = u3 - q * v3;

        u1 = v1;
        u2 = v2;
        u3 = v3;

        v1 = t1;
        v2 = t2;
        v3 = t3;
        z = 1;
    }
    uu = u1;
    vv = u2;

    if (vv < 0) {
          inverse = vv + PHI;
    } else {
          inverse = vv;
    }
    return inverse;
  }

  //finds out how many characters we can pack into n
  public int getNumEncrypChars(long n, long numOfAlpha)
  {
     return (int)(Math.log(n) / Math.log(numOfAlpha)) ;
  }
 
 //conver string to the long number since RSA is done 
 //in number. We can pack upto four chateracters to long number
  public long strToNum(String str)
  {

    int i=0; int j=0;
    long m=0;

    for ( i=str.length()-1; i>=0; i--)
    {
          m += str.charAt(i)*( Math.pow(128,i) );
    }
    return m;
  }

  //converts the long number to its corresponding string
  public String numToStr(long num, long numChar)
  {
     long mask = 127;
     long tmp;
     long num1 = num;
    
     StringBuffer sb = new StringBuffer(); 
     for ( int i=0; i< numChar; i++)
     { 
        num1 = num >> (7*i);
        tmp = num1 & mask ;
        sb.append( (char)tmp );
     }
     return sb.toString();
  }

//encrypts the numeric string
  public long encript(long M, long e, long n)
  {
   //we call decrypt here because both encrypt and decrypt are
   //basically the same modular arithmatics with different variables
    return descript(M, e,n);

  }

//decrypts the numeric string
  public long descript(long M, long d, long n)
  {
   
     int nbits = (int)(Math.ceil(Math.log(d)/ Math.log(2)));
     long OM =M;
      long UM=M;
//     int quot;
     
     BigDecimal   bdn = new BigDecimal( (double)n);
     BigDecimal   bdm = new BigDecimal( (double)UM);
     for ( int i=nbits-2; i>=0 ; i--)
     {
         if ( ( (d >> i) % 2 ) == 0 )
         {

/*              quot = (int) ( (double)(M*M) / (double) n );
              M = (M*M) - (quot*n); 
*/
             
             BigDecimal bdc = bdm.multiply( bdm );
             BigDecimal quot = bdc.divide( bdn,BigDecimal.ROUND_DOWN ); 
             bdm = bdc.subtract( quot.multiply(bdn)) ;

         }
         else
         {
/*
              quot = (int) ( (double)(M*M*OM) /  (double)n );
              M = (M*M*OM) - (quot*n); 
*/
             BigDecimal bdc = bdm.multiply( bdm );
             bdc = bdc.multiply( new BigDecimal( (double)OM) );
             BigDecimal quot = bdc.divide( bdn,BigDecimal.ROUND_DOWN ); 
             bdm = bdc.subtract( quot.multiply(bdn)) ;
         }
     }
     M =  (long)bdm.doubleValue() ;
     return M;
  }

  //encrypts the message
  public long[] encryptMsg( StringBuffer msg, long e, long n )
  {
        int nec = getNumEncrypChars(n, 128);

         //codeSize is the number of long  numbers we need to
         //encrypt the string
        int codeSize = (int)( Math.floor(msg.length()/nec) ) ;
        if ( (msg.length() % (int)nec) != 0 ) codeSize =
            (int)Math.floor(msg.length()/nec) +1;
        else
            codeSize = (int)Math.floor(msg.length()/nec) ;
          
        long codes[] = new long[codeSize];
        int j=0;

        msg.setLength(nec*codeSize);
        for ( int i=0; i<msg.length();i+=nec)
        {

           //breaks down the string into the substrings and encrypt them.
            String sstr = msg.toString().substring(i, i+nec) ;
            codes[j++] = encript(strToNum(sstr), e,n) ; 
        }
        return codes;
  }
//decrypt the message
  public String decryptMsg( long msg[], long d, long n )
  {
        int nec = getNumEncrypChars(n, 128);
        StringBuffer sb = new StringBuffer();
        int j=0;
        System.out.print("Before decrypting  : ");
        for ( int i=0; i<msg.length;i++)
        {
            System.out.print( numToStr(msg[i],nec) );
            sb.append(numToStr( descript(msg[i],d,n), nec));
        }
        System.out.println();
        System.out.println("After decrypting : "+sb.toString() );
        return sb.toString();
  }
  public static void main( String str[])
  {

    RSA rsa = new RSA();
    long p,q;
    p = 17569;
    q = 18401;
  //  p = 13;
   // q = 7;
    long n=p*q;
    long phi=(p-1)*(q-1);
    long na = 128;
    long i=0;

    long e = rsa.findE(p,q);    
    long d = rsa.extend( e, phi );
    long nec =  rsa.getNumEncrypChars(n,  na );

    System.out.println(e+" "+d+" "+nec+" "+n);
    long codes[] = rsa.encryptMsg(new StringBuffer(str[0]),e,n);
    System.out.println(rsa.decryptMsg(codes, d,n));
  }

}
